#include "qemu/osdep.h"
#include "qemu/error-report.h"
#include "qemu-common.h"
#include "qapi/error.h"
#include "hw/qdev.h"
#include "hw/block/block.h"
#include "sysemu/sysemu.h"
#include "sysemu/blockdev.h"
#include "sysemu/block-backend.h"
#include "libtcmu.h"

#include "block/aio.h"
#include "scsi/constants.h"

#define TYPE_TCMU_BLK "tcmu-blk"
#define TCMU_BLK(obj) OBJECT_CHECK(TcmuBlk, (obj), TYPE_TCMU_BLK)

typedef struct TcmuBlk {
    DeviceState parent;

    BlockConf conf;
} TcmuBlk;

typedef struct {
    struct tcmulib_cmd *cmd;
    struct tcmu_device *dev;
    QEMUIOVector qiov;
} TcmuBlkReq;

static TcmuBlkReq *tcmu_blk_req_new(struct tcmu_device *dev,
    struct tcmulib_cmd *cmd, struct iovec *iov, int niov)
{
    TcmuBlkReq *req = g_new(TcmuBlkReq, 1);

    req->dev = dev;
    req->cmd = cmd;
    qemu_iovec_init_external(&req->qiov, iov, niov);
    return req;
}

static void tcmu_blk_aio_cb(void *opaque, int ret)
{
    TcmuBlkReq *req = opaque;
    int result = ret ? CHECK_CONDITION : GOOD;

    tcmulib_command_complete(req->dev, req->cmd, result);
    tcmulib_processing_complete(req->dev);
    g_free(req);
}

#define ASCQ_INVALID_FIELD_IN_CDB 0x2400

static int tcmu_blk_handle_cmd(struct tcmu_device *dev,
    struct tcmulib_cmd *cmd)
{
    TcmuBlk *s = tcmu_get_dev_private(dev);
    BlockBackend *blk = s->conf.blk;
    uint8_t *cdb = cmd->cdb;
    uint64_t offset = tcmu_get_lba(cdb) << BDRV_SECTOR_BITS;
    TcmuBlkReq *req;

    switch (cdb[0]) {
    case INQUIRY:
        return tcmu_emulate_inquiry(dev, NULL, cdb, cmd->iovec,
            cmd->iov_cnt, cmd->sense_buf);
    case TEST_UNIT_READY:
        return tcmu_emulate_test_unit_ready(cdb, cmd->iovec,
            cmd->iov_cnt, cmd->sense_buf);
    case SERVICE_ACTION_IN_16:
        if (cdb[1] != SAI_READ_CAPACITY_16) {
            return TCMU_NOT_HANDLED;
        }
        return tcmu_emulate_read_capacity_16(blk_getlength(blk) / 512,
            512, cmd->cdb, cmd->iovec, cmd->iov_cnt, cmd->sense_buf);
    case MODE_SENSE:
    case MODE_SENSE_10:
        return tcmu_emulate_mode_sense(dev, cdb, cmd->iovec,
            cmd->iov_cnt, cmd->sense_buf);
    case MODE_SELECT:
    case MODE_SELECT_10:
        return tcmu_emulate_mode_select(dev, cdb, cmd->iovec,
            cmd->iov_cnt, cmd->sense_buf);

    case SYNCHRONIZE_CACHE:
    case SYNCHRONIZE_CACHE_16:
        if (cdb[1] & 0x2) {
            return tcmu_set_sense_data(cmd->sense_buf, ILLEGAL_REQUEST,
                ASCQ_INVALID_FIELD_IN_CDB, NULL);
        }
        req = tcmu_blk_req_new(dev, cmd, NULL, 0);
        blk_aio_flush(blk, tcmu_blk_aio_cb, req);
        return TCMU_ASYNC_HANDLED;
    case READ_6:
    case READ_10:
    case READ_12:
    case READ_16:
        req = tcmu_blk_req_new(dev, cmd, cmd->iovec, cmd->iov_cnt);
        blk_aio_preadv(blk, offset, &req->qiov, 0, tcmu_blk_aio_cb, req);
        return TCMU_ASYNC_HANDLED;
    case WRITE_6:
    case WRITE_10:
    case WRITE_12:
    case WRITE_16:
        req = tcmu_blk_req_new(dev, cmd, cmd->iovec, cmd->iov_cnt);
        blk_aio_pwritev(blk, offset, &req->qiov, 0, tcmu_blk_aio_cb, req);
        return TCMU_ASYNC_HANDLED;

    default:
        error_report("unknown command %x", cdb[0]);
        return TCMU_NOT_HANDLED;
    }
}

static void tcmu_blk_dev_event_handler(void *opaque)
{
    struct tcmu_device *dev = opaque;
    bool completed = false;

    tcmulib_processing_start(dev);

    for (;;) {
        struct tcmulib_cmd *cmd;
        int ret;

        cmd = tcmulib_get_next_command(dev);
        if (!cmd) {
            break;
        }

        ret = tcmu_blk_handle_cmd(dev, cmd);
        if (ret != TCMU_ASYNC_HANDLED) {
            tcmulib_command_complete(dev, cmd, ret);
            completed = true;
        }
    }

    if (completed) {
        tcmulib_processing_complete(dev);
    }
}

static int tcmu_blk_added(struct tcmu_device *dev)
{
    const char *cfgstring = tcmu_get_dev_cfgstring(dev);
    const char *name;
    BlockBackend *blk;
    TcmuBlk *s;

    name = strchr(cfgstring, '/');
    if (!name) {
        goto out;
    }

    blk = blk_by_name(name + 1);
    if (!blk) {
        goto out;
    }

    s = blk_get_attached_dev(blk);
    assert(s);

    tcmu_set_dev_private(dev, s);
    aio_set_fd_handler(blk_get_aio_context(s->conf.blk),
        tcmu_get_dev_fd(dev), true,
        tcmu_blk_dev_event_handler, NULL, NULL, dev);

out:
    return 0;
}

static void tcmu_blk_removed(struct tcmu_device *dev)
{
    TcmuBlk *s = tcmu_get_dev_private(dev);

    aio_set_fd_handler(blk_get_aio_context(s->conf.blk),
        tcmu_get_dev_fd(dev), false, NULL, NULL, NULL, NULL);
}

static void tcmu_blk_realize(DeviceState *dev, Error **errp)
{
    TcmuBlk *s = TCMU_BLK(dev);

    if (!s->conf.blk) {
        error_setg(errp, "drive property not set");
        return;
    }

    blkconf_apply_backend_options(&s->conf, blk_is_read_only(s->conf.blk),
        false, errp);
}

static Property tcmu_blk_properties[] = {
    DEFINE_BLOCK_PROPERTIES(TcmuBlk, conf),
    DEFINE_PROP_END_OF_LIST(),
};

static void tcmu_blk_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);

    dc->props = tcmu_blk_properties;
    dc->realize = tcmu_blk_realize;
}

static const TypeInfo tcmu_blk_info = {
    .name = TYPE_TCMU_BLK,
    .parent = TYPE_DEVICE,
    .instance_size = sizeof(TcmuBlk),
    .class_init = tcmu_blk_class_init,
};


#define TYPE_TCMU "tcmu"
#define TCMU(obj) OBJECT_CHECK(Tcmu, (obj), TYPE_TCMU)

typedef struct Tcmu {
    DeviceState parent;

    char *subtype;
    struct tcmulib_context *ctx;
    VMChangeStateEntry *vmchange;
} Tcmu;

static void tcmu_master_read(void *opaque)
{
    Tcmu *s = opaque;

    tcmulib_master_fd_ready(s->ctx);
}

static void tcmu_vm_state_change(void *opaque, int running, RunState state)
{
    Tcmu *s = opaque;

    if (running) {
        struct tcmulib_handler handler = {
            .hm_private = s,
            .subtype = s->subtype,
            .added = tcmu_blk_added,
            .removed = tcmu_blk_removed,
        };

        s->ctx = tcmulib_initialize(&handler, 1);
        if (!s->ctx) {
            error_setg(&error_abort, "failed to initialize tcmulib");
            return;
        }

        qemu_set_fd_handler(tcmulib_get_master_fd(s->ctx),
            tcmu_master_read, NULL, s);
        tcmulib_register(s->ctx);
    } else {
        if (!s->ctx) {
            return;
        }

        tcmulib_close(s->ctx);
        s->ctx = false;
    }
}

static void tcmu_realize(DeviceState *dev, Error **errp)
{
    Tcmu *s = TCMU(dev);

    if (!s->subtype) {
        error_setg(errp, "subtype is not defined");
        return;
    }

    s->vmchange = qemu_add_vm_change_state_handler(tcmu_vm_state_change, s);
}

static void tcmu_unrealize(DeviceState *dev, Error **errp)
{
    Tcmu *s = TCMU(dev);

    qemu_del_vm_change_state_handler(s->vmchange);
}

static Property tcmu_properties[] = {
    DEFINE_PROP_STRING("subtype", Tcmu, subtype),
    DEFINE_PROP_END_OF_LIST(),
};

static void tcmu_class_init(ObjectClass *klass, void *class_data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);

    dc->props = tcmu_properties;
    dc->realize = tcmu_realize;
    dc->unrealize = tcmu_unrealize;
}

static const TypeInfo tcmu_info = {
    .name = TYPE_TCMU,
    .parent = TYPE_DEVICE,
    .class_init = tcmu_class_init,
    .instance_size = sizeof(Tcmu),
};

static void tcmu_register_types(void)
{
    type_register_static(&tcmu_info);
    type_register_static(&tcmu_blk_info);
}

type_init(tcmu_register_types)
